
/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux

   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */



package sfclasses;

import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.applet.*;
import java.util.*;
import java.io.*;
import com.sun.java.swing.*;

/**
 * <b>Toplevel Frame of Remote Config Tool</b><br>
 * This class implements the main window of the control panel. The frame
 * is constructed when the user clicks on a button in the start page (see
 * sfRemote.java). It contains a drag and drop panel for the network
 * topology in the client area and provides access to all global 
 * configuration functions.
 * @version 1.0 12 Dec 1996
 * @author Roland E. Schmid
 */

public class ManageControl
	extends Frame
	implements ActionListener, ItemListener {
  /**
   * Initialize new main window<br>
   * The window is set up and a new ManageDomain object is created
   * holding the configuration data.
   * @param titleString Initial window title
   */
  public ManageControl(String titleString) {
    // window title bar
		super(titleString);
    if (titleString != null)
      setTitle(titleString);
    // create menus
    mbar = new MenuBar();
    FileMenu = new Menu("File");
	 MenuItem FileMenu1, FileMenu2, FileMenu3, FileMenu4, FileMenu5, FileMenu6, FileMenu7;
    FileMenu.add(FileMenu1 = new MenuItem("Load from Firewall"));
    FileMenu.add(FileMenu2 = new MenuItem("Save to Firewalls"));
    FileMenu.add(FileMenu3 = new MenuItem("Load from Local File"));
    FileMenu.add(FileMenu4 = new MenuItem("Save to Local File"));
    FileMenu.add(FileMenu5 = new MenuItem("Export Topology"));
    FileMenu.add(FileMenu6 = new MenuItem("Close"));
    FileMenu.addSeparator();
    FileMenu.add(FileMenu7 = new MenuItem("Quit"));
    mbar.add(FileMenu);

    EditMenu = new Menu("Edit"); 
	 MenuItem EditMenu1, EditMenu2, EditMenu3, EditMenu4, EditMenu5, EditMenu6;
    EditMenu.add(EditMenu1 = new MenuItem("General"));
    EditMenu.add(EditMenu2 = new MenuItem("Autoconfig"));
    EditMenu.add(EditMenu3 = new MenuItem("Rules"));
    EditMenu.add(EditMenu4 = new MenuItem("Notification Levels"));
    EditMenu.add(EditMenu5 = new MenuItem("Macros"));
    EditMenu.add(EditMenu6 = new MenuItem("Preferences"));
    EditMenu.addSeparator();
    EditMenu.add(TopologyCheckbox = new CheckboxMenuItem("Edit Topology", topologyEdit));
    mbar.add(EditMenu);

    ReportMenu = new Menu("Reports");
	 MenuItem ReportMenu1;
    ReportMenu.add(ReportMenu1 = new MenuItem("Configuration Report"));
    mbar.add(ReportMenu);

    HelpMenu = new Menu("Help");
	 MenuItem HelpMenu1, HelpMenu2;
    HelpMenu.add(HelpMenu1 = new MenuItem("User's Guide"));
    HelpMenu.add(HelpMenu2 = new MenuItem("About"));
    mbar.add(HelpMenu);
    mbar.setHelpMenu(HelpMenu);
    setMenuBar(mbar);

	 // register to receive events
	 FileMenu1.addActionListener(this);
	 FileMenu2.addActionListener(this);
	 FileMenu3.addActionListener(this);
	 FileMenu4.addActionListener(this);
	 FileMenu5.addActionListener(this);
	 FileMenu6.addActionListener(this);
	 FileMenu7.addActionListener(this);

	 EditMenu1.addActionListener(this);
	 EditMenu2.addActionListener(this);
	 EditMenu3.addActionListener(this);
	 EditMenu4.addActionListener(this);
	 EditMenu5.addActionListener(this);
	 EditMenu6.addActionListener(this);
	 TopologyCheckbox.addItemListener(this);	// CheckBoxMenu generates itemEvents
	 
	 ReportMenu1.addActionListener(this);

	 HelpMenu1.addActionListener(this);
	 HelpMenu2.addActionListener(this);

    // initialize data
    //appletContext = ac;
	 Communicator.parent = this;
    if (titleString != null) {
      byte host[] = Communicator.resolve(titleString);
      ServerInputStream ris = 
        new ServerInputStream(host, this.custom.getProperty("CONF_DIR") + this.custom.getProperty("GLOBAL_CONFIG"));
      if (ris.isEOF())
        mgDomain = null;
      else {
				PersistentInputStream is = new PersistentInputStream(ris);
				mgDomain = ManageDomain.loadDomainConfig(is);
      }
    }
    else
      mgDomain = new ManageDomain();
    if (mgDomain == null) { // host is unreachable
      prepareQuit();
      return; 
    }
    // create drag and drop panel
    try {
      hostClass = Class.forName("sfclasses.Host");
      netClass = Class.forName("sfclasses.Net");
      internetClass = Class.forName("sfclasses.Internet");
    }
    catch (ClassNotFoundException e) {
      System.out.println("Fatal error: class Host, Net or Internet not found!");
      terminated = true;
      return;
    }
    drawClass = hostClass;

    // create panels
    ddPanel = new DragDrop(mgDomain.getTopology(), drawClass, topologyEdit);
    Toolbar = new Panel();
    Toolbar.setLayout(new FlowLayout(FlowLayout.LEFT));

    CheckboxGroup g = new CheckboxGroup();
    Toolbar.add(hostBox = new Checkbox("Host", g, true));
    Toolbar.add(netBox = new Checkbox("Net", g, false));
    Toolbar.add(internetBox = new Checkbox("Internet", g, false));
	 	hostBox.addItemListener(this);
	 	netBox.addItemListener(this);
	 	internetBox.addItemListener(this);
    setLayout(new BorderLayout());
    add("Center", ddPanel);

    // set window title
    if (mgDomain.DomainName.length() > 0)
      setTitle(mgDomain.DomainName);
    else
      setTitle("Unnamed domain");

		ddPanel.checkConsistency(); 
		if (!(mgDomain.isConsistent())) 
			UserDialog.ErrorBox("Topology inconsistent!");

	 addWindowListener(new MCWAdapter());
	 UserDialog.WelcomeBox();
	 pack();
	 validate();
	 setSize(500, 350);
	 setVisible(true);
  }

	// Our own Adapter because we only process windowClosing events
  class MCWAdapter extends WindowAdapter {
  	public void windowClosing(WindowEvent we) {
      if (isEnabled())
        prepareQuit();
  	}
  } // class MCWAdapter

  public void actionPerformed(ActionEvent evt) {

  	 String arg = (String)evt.getActionCommand();

		// Process "Quit"
      if (arg.equals("Quit")) {
        prepareQuit();
      }

		// Process "Save to Firewalls"
      else if (arg.equals("Save to Firewalls")) {
        if (topologyEdit) {
          UserDialog.ErrorBox("Function not available while editing the topology!");
          return;
        }
        SaveData sd = new SaveData(mgDomain, this, false, false);
        setEnabled(false);
        setCursor(new Cursor(Cursor.WAIT_CURSOR));
        sd.start();
        return;
      }
		
		// Process "Save to Local File"
      else if (arg.equals("Save to Local File")) {
        if (topologyEdit) {
          UserDialog.ErrorBox("Function not available while editing the topology!");
          return;
        }
        SaveData sd = new SaveData(mgDomain, this, false, true);
        setEnabled(false);
        setCursor(new Cursor(Cursor.WAIT_CURSOR));
        sd.start();
        return;
      }

		// Process "Export Topology"
			else if (arg.equals("Export Topology")) {
				if (topologyEdit) {
					UserDialog.ErrorBox("Function not available while editing the topology!");
					return;
				}
        ExportData ed = new ExportData(this, mgDomain);
        setEnabled(false);
        setCursor(new Cursor(Cursor.WAIT_CURSOR));
        ed.start();
        return;
			}

		// Process "Load from Local File"
      else if (arg.equals("Load from Local File")) {
        if (topologyEdit) {
          UserDialog.ErrorBox("Function not available while editing the topology!");
          return;
        }
        LoadData ld = new LoadData(mgDomain, this, LoadData.LOCAL);
        setEnabled(false);
        setCursor(new Cursor(Cursor.WAIT_CURSOR));
        ld.start();
        return;
      }

		// Process "Load from Firewall"
      else if (arg.equals("Load from Firewall")) {
        if (topologyEdit) {
          UserDialog.ErrorBox("Function not available while editing the topology!");
          return;
        }
        LoadData ld = new LoadData(mgDomain, this, LoadData.REMOTE);
        setEnabled(false);
        setCursor(new Cursor(Cursor.WAIT_CURSOR));
        ld.start();
        return;
      }

		// Process "Close"
      else if (arg.equals("Close")) {
        LoadData ld = new LoadData(mgDomain, this, LoadData.NONE);
        setEnabled(false);
        setCursor(new Cursor(Cursor.WAIT_CURSOR));
        ld.start();
				Communicator.dispose();
        return;
      }

		// Process "General"
      else if (arg.equals("General")) {
        if (topologyEdit) {
          UserDialog.ErrorBox("Function not available while editing the topology!");
          return;
        }
        if (editRuleFrame != null && editRuleFrame.isShowing()) {
          UserDialog.ErrorBox("Function not available while editing rules!");
          return;
        }
        mgDomain.changed = true;
        mgDomain.autoconfValid = false;
        GeneralSettings gs;
        if (mgDomain.DomainName.length() == 0)
          gs = new GeneralSettings(this, "General Settings", mgDomain);
        else
          gs = new GeneralSettings(this, mgDomain.DomainName, mgDomain);
        gs.show();
        return;
      }

		// Process "Rules"
      else if (arg.equals("Rules")) {
        if (topologyEdit) {
          UserDialog.ErrorBox("Function not available while editing the topology!");
          return;
        }
        mgDomain.changed = true;
        if (editRuleFrame == null) {
          editRuleFrame = new RuleEditor(mgDomain);
          Dimension screensize = Toolkit.getDefaultToolkit().getScreenSize();
          editRuleFrame.setSize(screensize.width/6*5, screensize.height/4*3);
          editRuleFrame.setVisible(true);
        }
        else
          editRuleFrame.showAgain();
        if (!mgDomain.autoconfValid) {
          UserDialog.WarningBox("Autoconfig not executed. Rules may not be valid.");
        }
        return;
      }

		// Process "Autoconfig"
      else if (arg.equals("Autoconfig")) {
        if (topologyEdit) {
          UserDialog.ErrorBox("Function not available while editing the topology!");
          return;
        }
        if (editRuleFrame != null && editRuleFrame.isShowing()) {
          UserDialog.ErrorBox("Function not available while editing rules!");
          return;
        }
        mgDomain.changed = true;
        AutoconfThread at = new AutoconfThread(mgDomain, this);
        setEnabled(false);
        setCursor(new Cursor(Cursor.WAIT_CURSOR));
        at.start();
        return;
      }

		// Process "Notification Levels"
      else if (arg.equals("Notification Levels")) {
        NotificationSelect notificationSelect = new NotificationSelect(mgDomain, this);
        notificationSelect.setVisible(true);
        return;
      }

      else if (arg.equals("Macros")) {
        MacroSelect macroSelect = new MacroSelect(mgDomain, this);
        macroSelect.setVisible(true);
        return;
      }

      else if (arg.equals("Preferences")) {
				Preferences p = new Preferences();
				p.setVisible(true);
        return;
      }

		// Process "Configuration Report"
      else if (arg.equals("Configuration Report")) {
        ReportThread rt = new ReportThread(mgDomain, this, ReportThread.REPORT_CONFIG);
        setEnabled(false);
        setCursor(new Cursor(Cursor.WAIT_CURSOR));
        rt.start();
        return;
      }

		// Process "User's Guide"
      else if (arg.equals("User's Guide")) {
			UserDialog.NoticeBox("Documentation only available as a PostScript file.");
      return;
      }

		// Process "About"
      else if (arg.equals("About")) {
				
				AboutDialog ad = new AboutDialog(this);
				ad.setVisible(true);
				return;
      }
	} // end actionPerformed



	public void itemStateChanged(ItemEvent ie) {

	String arg = (String)ie.getItem();

		// Process "Edit Topology"
     if (arg.equals("Edit Topology")) {
        if (editRuleFrame != null && editRuleFrame.isShowing()) {
          UserDialog.ErrorBox("Function not available while editing rules!");
          topologyEdit = false;
          TopologyCheckbox.setState(topologyEdit);
          return;
        }
        topologyEdit = TopologyCheckbox.getState();
        mgDomain.changed = true;
        mgDomain.autoconfValid = false;
        if (topologyEdit) {
          ddPanel.enableEdit();
          add("North", Toolbar);
          validate();
        }
        else {
          if (!ddPanel.disableEdit()) {
            topologyEdit = true;
            TopologyCheckbox.setState(topologyEdit);
          }
          remove(Toolbar);
          mgDomain.topologyChanged();
          validate();
					if (mgDomain.isConsistent()) {
          	UserDialog.NoticeBox("Run autoconfiguration now to update the rules!");
					}
					else {
						UserDialog.ErrorBox("Topology is inconsistent. Please fix prior to running autoconfiguration!");
					}
        }
        return;
		}

		// Process Checkbox "Host"
		 else if (arg.equals("Host")) {
			drawClass = hostClass;
			ddPanel.setClass(drawClass);
			return;
		 }

		// Process Checkbox "Net"
		 else if (arg.equals("Net")) {
			drawClass = netClass;
			ddPanel.setClass(drawClass);
			return;
		 }

		// Process Checkbox "Internet"
		 else if (arg.equals("Internet")) {
			drawClass = internetClass;
			ddPanel.setClass(drawClass);
			return;
		 }
	} 		// itemStateChanged


  protected void prepareQuit() {
    if (mgDomain != null && mgDomain.changed) {
      SaveData sd = new SaveData(mgDomain, this, true, false);
      setEnabled(false);
      setCursor(new Cursor(Cursor.WAIT_CURSOR));
      sd.start();
      return;
    }     

    /* cleanup stuff */
    Communicator.dispose();
		setVisible(false);
    dispose();
    ddPanel = null;
    mgDomain = null;
		System.exit(0);
  }

  protected void setManageDomain(ManageDomain newmd) {
    remove(ddPanel);
    ddPanel = new DragDrop(newmd.getTopology(), drawClass, topologyEdit);
    add("Center", ddPanel);
		ddPanel.checkConsistency();
    validate();
		if (!newmd.isConsistent()) {
			UserDialog.ErrorBox("Topology is inconsistent!");
		}

    if (newmd.DomainName.length() > 0)
      setTitle(newmd.DomainName);
    else
      setTitle("Unnamed domain");

    mgDomain = newmd;
  }

  // public data
  public boolean terminated = false;
	public static Custom custom;

  // private data

  // gui data
  private DragDrop ddPanel;
  private MenuBar mbar;
  private Menu FileMenu;
  private Menu EditMenu;
  private Menu ReportMenu;
  private Menu HelpMenu;
  private CheckboxMenuItem TopologyCheckbox;
  private Panel Toolbar;
  private Checkbox hostBox;
  private Checkbox netBox;
  private Checkbox internetBox;
  private RuleEditor editRuleFrame;

  // application data
  private ManageDomain mgDomain;
  private boolean topologyEdit = false;
  private Class drawClass, hostClass, netClass, internetClass;

} // ManageControl


class ExportData extends Thread {

	public ExportData(Frame p, ManageDomain mgDomain) {
	parent = p;
	md = mgDomain;
	}

	public void run() {

	Graph g;
	Enumeration en;
	AddrList adr;
	ManageObject mo;
	int ID;
	Host host;
	Net net;
	
		FileDialog fd = new FileDialog(parent, "Export Topology", FileDialog.SAVE);
		fd.setVisible(true);

		if (fd.getFile() == null) {
			parent.setEnabled(true);
			parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
			return;
		}
		String filename = fd.getDirectory() + fd.getFile();

		BusyBox b = new BusyBox("Save in progress. Please wait!");
		b.start(); // show busy box, returns immediately
		try {
			FileOutputStream fos = new FileOutputStream(filename);

			pw = new PrintWriter(fos);
			Date date = new Date();

			pw.println("#");
			pw.println("# Topology Export");
			pw.println("# Generated on " + date + "by sfControl");
			pw.println("#");
			pw.println();

			g = md.getTopology();
			en = g.getAllVertices();
			ID = 0;

			while (en.hasMoreElements()) {
			
				ID += 1;
				mo = (ManageObject)en.nextElement();
				mo.uniqueID = ID;

				pw.println("$item");
				pw.println("id=" + ID);
				pw.print("type=");

				// process host
				host = null;
				if (mo instanceof Host) {
					host = (Host)mo;
					if (host.isFirewall & host.isServer)
						pw.println("(firewall, server)");
					else
						if (host.isFirewall)
							pw.println("firewall");
						if (host.isServer)
							pw.println("server");

					adr = host.HostAddresses;
					writeAddresses(adr);
					if (host.isServer)
						for (int i = 0; i < host.hostTemplates.length; i++)
							if (host.hostTemplates[i])
								pw.println("attribute=(service, \""+Templates.prompt[Templates.startServer+i]+"\")");

				}

				// process net
				net = null;
				if (mo instanceof Net) {
					net = (Net)mo;
					pw.println("net");
					adr = net.NetAddresses;
					writeAddresses(adr);
				}

				// process internet
				if (mo instanceof Internet) {
					pw.println("internet");
				}

				pw.println("$enditem");
				pw.println();

			}

			// export connections

			pw.println("# Follows the adjazenzlists of all connections.");
			pw.println("# Numbers refer to the item's id.");
			pw.println("$connections");

			en = g.getAllVertices();
			Enumeration neighbourEnum;
			while (en.hasMoreElements()) {

				mo = (ManageObject)en.nextElement();
				pw.print("neighbours{" + mo.uniqueID + "}=(");
				
				neighbourEnum = g.findNeighbors(mo);
				if (neighbourEnum.hasMoreElements()) {
					mo = (ManageObject)neighbourEnum.nextElement();
					pw.print(mo.uniqueID);
				
					while (neighbourEnum.hasMoreElements()) {
						mo = (ManageObject)neighbourEnum.nextElement();
						pw.print(", " + mo.uniqueID);
					}
					pw.println(")");
				}
				else
					pw.println("()");
			}
			pw.println("$endconnections");

			pw.flush();
			pw.close();
			fos.close();
			b.done();


			parent.setEnabled(true);
			parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));

		} // try
		catch (IOException e) {
			b.done();
			UserDialog.ErrorBox("Error saving data to "+filename+". Data not saved! ("+Communicator.commError+")");
		} // catch
	}

	
	private void writeAddresses(AddrList al) {

	AddrObj ao;
	String s;
	byte[] address;
	byte[] netmask;

		Vector av = al.getAddresses();
		
		for (int i = 0; i < av.size(); i++) {
			ao = (AddrObj)av.elementAt(i);
			address = ao.getAddr();
			netmask = ao.getMask();

			if ((address == null) || (address.length != 4))
				pw.print("address=(0.0.0.0");
			else {
				s = ("" + Utils.unsign(address[0]));
				for (i=1; i < 4; i++)
					s += ("." + Utils.unsign(address[i]));
				pw.print("address=(" + s);
			}
			if ((netmask == null) || (netmask.length != 4))
				pw.println("/0.0.0.0)");
			else {
				s = ("" + Utils.unsign(netmask[0]));
				for (i=1; i < 4; i++)
					s += ("." + Utils.unsign(netmask[i]));
				pw.println("/" + s + ")");
			}
		}	
	}
	
	
protected Frame parent;
protected ManageDomain md;
protected PrintWriter pw;
}

// Thread used for saving data. A separate thread is necessary, because
// it is not possible to wait for the user response when showing a dialog
// box from the event handling loop.
class SaveData extends Thread {

  public SaveData(ManageDomain md, ManageControl p, boolean ask, boolean sl) {
    mgDomain = md;
    parent = p;
    askDialog = ask;
    saveLocal = ask ? false : sl;
  }

  public void run() {

    if (askDialog) {
			int answer = JOptionPane.showConfirmDialog(null,
				"Save data to firewalls before exiting?",
				"Data not saved",
				JOptionPane.YES_NO_CANCEL_OPTION);
      parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
      switch (answer) {
        case JOptionPane.YES_OPTION: // save and quit
          break;
        case JOptionPane.NO_OPTION: // don't save, quit
          mgDomain.changed = false;
          parent.prepareQuit();
          return;
        case JOptionPane.CANCEL_OPTION: // don't save, don't quit
          parent.setEnabled(true);
          return;
      }
      parent.setCursor(new Cursor(Cursor.WAIT_CURSOR));
    }

    boolean success = true;
    if (!saveLocal) { // save data to firewalls
      BusyBox b = new BusyBox("Save in progress. Please wait!");
      b.start(); // show busy box, returns immediately
      Enumeration en = mgDomain.Firewalls.elements();
      Host f;
      byte faddr[];
      ServerOutputStream ros;
      PersistentOutputStream os;
      while (en.hasMoreElements()) {
        f = (Host)en.nextElement();
        faddr = f.HostAddresses.getFirstAddress();
        try {
          ros = new ServerOutputStream(faddr, ManageControl.custom.getProperty("GLOBAL_CONFIG"));
          if (ros.isClosed()) {
            throw new IOException();
					}
					os = new PersistentOutputStream(ros);
          if (!mgDomain.saveDomainConfig(os))
            throw new IOException();
          ros.close();
          os.close();
        } // try
        catch (IOException e) {
          UserDialog.ErrorBox("Error saving data to "+f.objectID+" ("+Utils.printIP(faddr)+").");
          success = false;
        } // catch
      } // while
      b.done();
    } // if
    else { // save data to local file
			FileDialog fd = new FileDialog(parent, "Save Configuration", FileDialog.SAVE);
			fd.setVisible(true);
			if (fd.getFile() == null) {
				parent.setEnabled(true);
				parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
				return;
			}
			String filename = fd.getDirectory() + fd.getFile();
      BusyBox b = new BusyBox("Save in progress. Please wait!");
      b.start(); // show busy box, returns immediately
      try {
        FileOutputStream fos = new FileOutputStream(filename);
        PersistentOutputStream os = new PersistentOutputStream(fos);
        if (!mgDomain.saveDomainConfig(os))
          throw new IOException("Error writing data");
        fos.close();
        os.close();
        b.done();
      } // try
      catch (IOException e) {
        b.done();
        UserDialog.ErrorBox("Error saving data to "+filename+". Data not saved! ("+Communicator.commError+")");
        success = false;
      } // catch
    }

    if (success) {
      mgDomain.changed = false;
      UserDialog.NoticeBox("Data saved successfully!");
      if (askDialog) // quit after saving
        parent.prepareQuit();
    }

    parent.setEnabled(true);
    parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
  }

  private ManageDomain mgDomain;
  private ManageControl parent; // parent frame
  private boolean askDialog, saveLocal;

} // SaveData 

// This thread is used for loading data
class LoadData extends Thread {

  public LoadData(ManageDomain md, ManageControl p, int l) {
    mgDomain = md;
    parent = p;
    load = l;
  }

  public void run() {

    if (mgDomain.changed) {
			int answer = JOptionPane.showConfirmDialog(null,
				"Save data to firewalls?",
				"Data not saved",
				JOptionPane.YES_NO_CANCEL_OPTION);
      switch (answer) {
        case JOptionPane.YES_OPTION: // save and load
          SaveData sd = new SaveData(mgDomain, parent, false, false);
          sd.start();
          while (sd.isAlive()) {
            try {
              sd.join(); // wait for save thread to complete
              break;
            }
            catch (Exception ignore) {}
          }
          break;
        case JOptionPane.NO_OPTION: // don't save, load break;
					break;
        case JOptionPane.CANCEL_OPTION: // don't save, abort load
          parent.setEnabled(true);
          parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
          return;
      }
    }

    if (load == NONE) 
      parent.setManageDomain(new ManageDomain());

    if (load == REMOTE) {
      StringDialog strd = new StringDialog(parent, "Hostname or IP address:");
      strd.show();
      if (strd.getAnswer() != StringDialog.OK) {
        parent.setEnabled(true);
        parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
        return;
      }
      String hostname = strd.getString();
      byte host[] = Communicator.resolve(hostname);
      if (host == null) {
        switch (Communicator.commError) {
          case Communicator.COMM_RESOLVE:
            UserDialog.ErrorBox("Unknown host "+hostname);
            break;
          default:
            UserDialog.ErrorBox("Error while checking hostname. ("+Communicator.commError+")");
            break;
        }
      }
      else {
        ManageDomain newmd;
        ServerInputStream ris =
          new ServerInputStream(host, ManageControl.custom.getProperty("CONF_DIR") + ManageControl.custom.getProperty("GLOBAL_CONFIG"));
        if (ris.isEOF())
          newmd = null;
        else {
          PersistentInputStream is = new PersistentInputStream(ris);
					newmd = ManageDomain.loadDomainConfig(is);
        }

        if (newmd == null) { // error reading file
          UserDialog.ErrorBox("Error loading data! ("+Communicator.commError+")");
        }
        else
          parent.setManageDomain(newmd);
      }
    }

    if (load == LOCAL) {
			FileDialog fd = new FileDialog(parent, "Load configuration", FileDialog.LOAD);
			fd.setVisible(true);
			String filename = fd.getDirectory();
			if (fd.getFile() == null) {
				parent.setEnabled(true);
				parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
				return;
			}
			filename += fd.getFile();

			FileInputStream fis = null;
			try {
				fis = new FileInputStream(filename);
			}
			catch (FileNotFoundException e) {
			}	
      ManageDomain newmd;
      if (fis == null)
        newmd = null;
      else {
				PersistentInputStream is = new PersistentInputStream(fis);
				newmd = ManageDomain.loadDomainConfig(is);
      }
      if (newmd == null) // error reading file
				UserDialog.ErrorBox("Can't open File: "+filename);
      else
        parent.setManageDomain(newmd);
    }

    parent.setEnabled(true);
    parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
  }

  private ManageDomain mgDomain;
  private ManageControl parent; // parent frame
  private int load; // if false create empty domain

  public static final int NONE = 0;
  public static final int LOCAL = 1;
  public static final int REMOTE = 2;

} // LoadData

// This thread executes autoconfiguration
class AutoconfThread extends Thread {

  public AutoconfThread(ManageDomain md, Frame p ) {
    mgDomain = md;
    parent = p;
  }

  public void run() {
    BusyBox bb = new BusyBox("Running Autoconfiguration. Please wait!");
    bb.start();
    try {
      sleep(500); // time to show busy box
    }
    catch (Exception e) {}
    mgDomain.autoConfig();
    bb.done();
    UserDialog.NoticeBox("Autoconfiguration completed successfully!");
    parent.setEnabled(true);
    parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
  }

  private ManageDomain mgDomain;
  private Frame parent;
} // AutoconfThread

// This thread generates and shows reports
class ReportThread extends Thread {

  public ReportThread(ManageDomain md, Frame p, int rt) {
    mgDomain = md;
    parent = p;
    reportType = rt;
  }

  public void run() {
    BusyBox bb = new BusyBox("Generating report. Please wait!");
    bb.start();
		String filename = ManageControl.custom.getProperty("TMP_DIR") + "ConfigReport.html";
    if (filename == null) {
      UserDialog.ErrorBox("Cannot create temporary file name! ("+Communicator.commError+")");
      bb.done();
      parent.setEnabled(true);
      parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
      return;
    }
 
    try {

      FileOutputStream fos = new FileOutputStream(filename, false);
      PrintWriter pw = new PrintWriter(fos, true);

      switch (reportType) {
        case REPORT_CONFIG:
          writeConfigReport(pw);
          break;
        default:
          UserDialog.ErrorBox("Unknown report type!");
          bb.done();
          parent.setEnabled(true);
          parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
          return;
      }
      fos.close();
      pw.close();

			String URLString = Utils.fileURLPrefix() + filename;
			if (System.getProperty("os.name").equals("Windows 95")) {
				URLString.replace('\\', '/');
			}

			URL	u = new URL(URLString);
			Frame hv = new HTMLView(u);
    }
    catch (IOException e) {
      UserDialog.ErrorBox("Error writing file! ("+Communicator.commError+")");
    }
    finally {
      bb.done();
    parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
    parent.setEnabled(true);
    }
  }

  /**
   * generate configuration report
   */
  private void writeConfigReport(PrintWriter ps) {
    ps.println("<html><head><title>sf Configuration Report</title></head><body>");
    ps.println("<h1>sf Configuration Report</h1>");
    ps.println("<p>Report generated "+(new Date()).toString());
    if (mgDomain.DomainName.length() > 0)
      ps.println("<h2>Management domain <i>"+mgDomain.DomainName+"</i></h2>");

    Enumeration en = mgDomain.getTopology().getAllVertices();
    ManageObject mo;
    Host h;
    Net n;
    Internet in;
    while (en.hasMoreElements()) {
      mo = (ManageObject)en.nextElement();
      if (mo instanceof Host) {
        h = (Host)mo;
        ps.println("<h3><a name="+mo.identString()+">Host <i>"+h.objectID+"</i></a></h3>");
        ps.println("<ul>");
        if (h.level == 1)
          ps.println("<li>Level: "+h.level+" firewall hop to nearest Internet connection");
        else if (h.level >= 0)
          ps.println("<li>Level: "+h.level+" firewall hops to nearest Internet connection");
        else
          ps.println("<li>No Internet connection");
        if (h.isFirewall)
          if (h.transparent)
            ps.println("<li>Firewall (allows packets from outside to outside)");
          else
            ps.println("<li>Firewall (blocks packets from outside to outside)");
        if (h.isServer && h.publicServer)
          ps.println("<li>Server (public access)");
        if (h.isServer && !h.publicServer)
          ps.println("<li>Server (internal access only)");
        ps.println("<li>Addresses:");
        writeAddresses(ps, h.HostAddresses);
        if (h.isFirewall) {
          ps.println("<li>Internalnets:");
          writeAddresses(ps, h.internalAddresses);
        }
        writeNeighbors(ps, mo);
        ps.println("</ul>");
      }
      else if (mo instanceof Net) {
        n = (Net)mo;
        ps.println("<h3><a name="+mo.identString()+">Network <i>"+n.objectID+"</i></a></h3>");
        ps.println("<ul>");
        if (n.level == 1)
          ps.println("<li>Level: "+n.level+" firewall hop to nearest Internet connection");
        else if (n.level >= 0)
          ps.println("<li>Level: "+n.level+" firewall hops to nearest Internet connection");
        else
          ps.println("<li>No Internet connection");
        ps.println("<li>Addresses:");
        writeAddresses(ps, n.NetAddresses);
        writeNeighbors(ps, mo);
        ps.println("</ul>");
      }
      else if (mo instanceof Internet) {
        in = (Internet)mo;
      }
      else {
        ps.println("<h3><a name="+mo.identString()+">Unknown object: <i>"+mo+"</i></a></h3>");
      }
    }

    ps.println("</body></html>");
  }

  private void writeAddresses(PrintWriter ps, AddrList al) {
    if (al==null)
      return;
    Enumeration en = al.getAddresses().elements();
    AddrObj ao;
    ps.println("<ul>");
    while (en.hasMoreElements()) {
      ao = (AddrObj)en.nextElement();
      ps.println("<li><tt>"+ao.toString()+"</tt>");
    }
    ps.println("</ul>");
  }

  private void writeNeighbors(PrintWriter ps, ManageObject mo) {
    Enumeration en = mgDomain.getTopology().findNeighbors(mo);
    ManageObject n;
    ps.println("<li>Connected to:<ul>");
    while (en.hasMoreElements()) {
      n = (ManageObject)en.nextElement();
      if (n.objectID.length() > 0)
        if (n instanceof Internet)
          ps.println("<li>"+n.objectID);
        else
          ps.println("<li><a href=\"#"+n.identString()+"\">"+n.objectID+"</a>");
      else if (n instanceof Host)
        ps.println("<li><a href=\"#"+n.identString()+"\">Unnamed host"+"</a>");
      else if (n instanceof Net)
        ps.println("<li><a href=\"#"+n.identString()+"\">Unnamed network"+"</a>");
      else
        ps.println("<li><a href=\"#"+n.identString()+"\">Unnamed object"+"</a>");
    }
    ps.println("</ul>");
  }

  public static final int REPORT_CONFIG = 1; // configuration report

  private ManageDomain mgDomain;
  private Frame parent;
  private BusyBox busybox;
  private int reportType;
} // ReportThread

// This dialog box is used to enter the general access profile
// and the configuration client addresses
class GeneralSettings
	extends Dialog
	implements ActionListener, TextListener, ItemListener {

  public GeneralSettings(Frame p, String title, ManageDomain md) {
    super(p, title, true);
    parent = p;
    mgDomain = md;
    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.insets = new Insets(5,5,5,5);
    setLayout(gbl);

    // Name panel
    Utils.add_component(this, new Label("Name: "), gbl, gbc, 0, 0, 1, 1, 0, 0);
    DNameText = new TextField(mgDomain.DomainName);
	 DNameText.addTextListener(this);
    Utils.add_component(this, DNameText, gbl, gbc, 1, 0, 3, 1, 100, 0);

    // Default mail addresses
    Utils.add_component(this, new Label("Default mail:"),
	 													gbl, gbc, 0, 1, 1, 1, 0, 0);
    mailText = new TextField(mgDomain.defaultMail);
	 mailText.addTextListener(this);
    Utils.add_component(this, mailText, gbl, gbc, 1, 1, 3, 1, 100, 0);

    int half = (int)Math.floor(((double)Templates.numberGeneral / 2.0) + 0.5);

    // Address Panel (configuration clients)
    addrPanel = mgDomain.configClients.editPanel();
    Utils.add_component(this, new Label("Configuration clients:"),
	 													gbl, gbc, 0, 2, 2, 1, 0, 0);
    Utils.add_component(this, addrPanel, gbl, gbc, 0, 3, 2, half-1, 100, 100);

    // Checkboxes
    gbc.fill = GridBagConstraints.NONE;
    gbc.anchor = GridBagConstraints.NORTHWEST;
    CheckboxArray = new Checkbox[Templates.numberGeneral];
    for (int i=0; i < Templates.numberGeneral; i++) {
      CheckboxArray[i] = new Checkbox(Templates.prompt[i],
													mgDomain.generalTemplates[i]);
      CheckboxArray[i].addItemListener(this);

      if (i < half)
        Utils.add_component(this, CheckboxArray[i], gbl, gbc, 
                            2, i+2, 1, 1, 0, 0);
      else
        Utils.add_component(this, CheckboxArray[i], gbl, gbc, 
                            3, i+2-half, 1, 1, 0, 0);
    }

    // Buttons
    OKbutton = new Button("OK");
	 OKbutton.addActionListener(this);
    Utils.add_component(this, OKbutton, gbl, gbc, 0, 2+half, 4, 1, 0, 0);

    // Window size
    pack();
	 addWindowListener(new GSAdapter());
  } // GeneralSettings

  class GSAdapter extends WindowAdapter {
  	public void windowClosing(WindowEvent we) {
      if (isEnabled())
        dispose();
	}
  } // Class GSAdapter

	// Eventhandling routines
  public void textValueChanged(TextEvent te) {

  Object source = te.getSource();
  	if (source == DNameText) {
      mgDomain.DomainName = (DNameText.getText().trim());
      setTitle(mgDomain.DomainName);
      parent.setTitle(mgDomain.DomainName);
      return;
		}
	} // textValueChanged


  public void actionPerformed(ActionEvent ae) {

  Object source = ae.getSource();
	 if (source == OKbutton) {
		mgDomain.DomainName = DNameText.getText().trim();
		if (mgDomain.DomainName.length() > 0)
		  parent.setTitle(mgDomain.DomainName);
		dispose();
		return;
    }
	} // actionPerformed


	public void itemStateChanged(ItemEvent ie) {

	Object source = ie.getSource();
   if (source instanceof Checkbox) {
      for (int i=0; i < Templates.numberGeneral; i++) 
        if (source==CheckboxArray[i])
          mgDomain.generalTemplates[i] = CheckboxArray[i].getState();
      return;
    }
  } // itemStateChanged


  private ManageDomain mgDomain;

  private Frame parent;

	private TextField DNameText;
	private TextField mailText;
	private Panel addrPanel;
	private Checkbox CheckboxArray[]; 
	private Button OKbutton;

} // Class GeneralSettings

// This dialog box provides a selection list for editing notification levels
class NotificationSelect
	extends Dialog
	implements Refreshable, ActionListener {

  public NotificationSelect(ManageDomain md, ManageControl p) {
    super(p, "Notification Levels", true);

    p.setCursor(new Cursor(Cursor.WAIT_CURSOR));

    mgDomain = md;
    parent = p;

    gbl = new GridBagLayout();
    gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.insets = new Insets(3,3,3,3);
    setLayout(gbl);

    nList = new List(15, false);
    Notification.constructList(nList, mgDomain.NLevels);
    OKbutton = new Button("OK");
    AddButton = new Button("Add new level");
    DeleteButton = new Button("Delete");
    ChangeButton = new Button("Change");

	 OKbutton.addActionListener(this);
	 AddButton.addActionListener(this);
	 DeleteButton.addActionListener(this);
	 ChangeButton.addActionListener(this);

    Utils.add_component(this, nList, gbl, gbc, 0, 0, 3, 15, 100, 100);

    Utils.add_component(this, OKbutton, gbl, gbc, 3, 0, 1, 1, 0, 0);
    Utils.add_component(this, AddButton, gbl, gbc, 3, 1, 1, 1, 0, 0);
    Utils.add_component(this, DeleteButton, gbl, gbc, 3, 2, 1, 1, 0, 0);
    Utils.add_component(this, ChangeButton, gbl, gbc, 3, 3, 1, 1, 0, 0);
    pack();

    p.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
  }

  public void actionPerformed(ActionEvent ae) {

  Object source = ae.getSource();

	// Process OKbutton
    if (source == OKbutton) {
      dispose();
      return;
    }

	// Process AddButton
    else if (source == AddButton) {
      Notification ln;
      try {
        ln = (Notification)mgDomain.NLevels.lastElement();
      }
      catch (NoSuchElementException e) {
        ln = null;
      }
      Notification not = new Notification
        (ln == null ? Notification.firstDynamic : ln.levelNumber + 1);
      mgDomain.NLevels.addElement(not);
      not.edit(this, parent, mgDomain, true);
      return;
    }

	// Process DeleteButton
    else if (source == DeleteButton) {
      String str = nList.getSelectedItem();
      if (str == null)
        return;
      Notification not = Notification.getLevel(str, mgDomain.NLevels);
      if (not == null)
        return;
      if (not.levelNumber < Notification.firstDynamic) {
        UserDialog.ErrorBox("Cannot delete static notification level!");
        return;
      }
// !!! this doesn't work, cause we are in the event loop
//      OkCancelBox oc = new OkCancelBox(parent, "Deleting notification level "+str+".");
//      oc.show();
//      if (oc.getAnswer() != OkCancelBox.OK)
//        return true;
      mgDomain.NLevels.removeElement(not);
      refresh();
      return;
    }
	 
	 // Process ChangeButton
    else if (source == ChangeButton) {
      String str = nList.getSelectedItem();
      if (str == null)
        return;
      Notification not = Notification.getLevel(str, mgDomain.NLevels);
      if (not == null)
        return;
      not.edit(this, parent, mgDomain, false);
      return;
    }
  } // actionPerformed


  public void refresh() {
    parent.setCursor(new Cursor(Cursor.WAIT_CURSOR));
    remove(nList);
    nList = new List(15, false);
    Notification.constructList(nList, mgDomain.NLevels);
    Utils.add_component(this, nList, gbl, gbc, 0, 0, 3, 15, 100, 100);
    validate();
    parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
  }

  private ManageDomain mgDomain;
  private ManageControl parent;
 
  private GridBagLayout gbl;
  private GridBagConstraints gbc;

  private List nList;
  private Button OKbutton;
  private Button AddButton;
  private Button DeleteButton;
  private Button ChangeButton;

} // NotificationSelect

// This dialog box implements a selection list for editing macros
class MacroSelect
	extends Dialog
	implements Refreshable, ActionListener {

  public MacroSelect(ManageDomain md, ManageControl p) {
    super(p, "Address and Service Macros", true);

    p.setCursor(new Cursor(Cursor.WAIT_CURSOR));

    mgDomain = md;
    parent = p;

    gbl = new GridBagLayout();
    gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.insets = new Insets(3,3,3,3);
    setLayout(gbl);

    nList = new List(15, false);
    Macro.constructList(nList, mgDomain.Macros);
    OKbutton = new Button("OK");
    AddButton = new Button("Add new macro");
    DeleteButton = new Button("Delete");
    ChangeButton = new Button("Change");

	 OKbutton.addActionListener(this);
	 AddButton.addActionListener(this);
	 DeleteButton.addActionListener(this);
	 ChangeButton.addActionListener(this);

    Utils.add_component(this, nList, gbl, gbc, 0, 0, 3, 15, 100, 100);
    Utils.add_component(this, OKbutton, gbl, gbc, 3, 0, 1, 1, 0, 0);
    Utils.add_component(this, AddButton, gbl, gbc, 3, 1, 1, 1, 0, 0);
    Utils.add_component(this, DeleteButton, gbl, gbc, 3, 2, 1, 1, 0, 0);
    Utils.add_component(this, ChangeButton, gbl, gbc, 3, 3, 1, 1, 0, 0);

    pack();

    p.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
  }

  public void actionPerformed(ActionEvent ae) {
  Object source = ae.getSource();

	// Process OKbutton
    if (source == OKbutton) {
      dispose();
      return;
    }

	 // Process AddButton
    else if (source == AddButton) {
      Macro ln;
      try {
        ln = (Macro)mgDomain.Macros.lastElement();
      }
      catch (NoSuchElementException e) {
        ln = null;
      }
      Macro m = new Macro
        (ln == null ? Macro.firstDynamic : ln.macroNumber - 1);
      mgDomain.Macros.addElement(m);
      m.edit(this, parent, mgDomain, true);
      return;
    }

	 // Process DeleteButton
    else if (source == DeleteButton) {
      String str = nList.getSelectedItem();
      if (str == null)
        return;
      Macro m = Macro.getMacro(str, mgDomain.Macros);
      if (m == null)
        return;
      if (m.macroNumber > Macro.firstDynamic) {
        UserDialog.ErrorBox("Cannot delete static macro!");
        return;
      }
// !!! this doesn't work, cause we are in the event loop
//      OkCancelBox oc = new OkCancelBox(parent, "Deleting macro "+str+".");
//      oc.show();
//      if (oc.getAnswer() != OkCancelBox.OK)
//        return true;
      mgDomain.Macros.removeElement(m);
      refresh();
      return;
    }

	 // Process ChangeButton
    else if (source == ChangeButton) {
      String str = nList.getSelectedItem();
      if (str == null)
        return;
      Macro m = Macro.getMacro(str, mgDomain.Macros);
      if (m == null)
        return;
      m.edit(this, parent, mgDomain, false);
      return;
    }
  } // actionPerformed



  public void refresh() {
    parent.setCursor(new Cursor(Cursor.WAIT_CURSOR));
    remove(nList);
    nList = new List(15, false);
    Macro.constructList(nList, mgDomain.Macros);
    Utils.add_component(this, nList, gbl, gbc, 0, 0, 3, 15, 100, 100);
    validate();
    parent.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
  }

  private ManageDomain mgDomain;
  private ManageControl parent;
 
  private GridBagLayout gbl;
  private GridBagConstraints gbc;

  private List nList;
  private Button OKbutton;
  private Button AddButton;
  private Button DeleteButton;
  private Button ChangeButton;

} // MacroSelect
