/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux

   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */


/*
 * Firewall-to-firewall communication
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "sf_custom.h"
#include "sf_control.h"
#include "sf_global.h"
#include "sf_config.h"
#include "sf_daemon.h"
#include "sf_comm.h"

int sockcomm    = -1;
int sockfriends = -1;

void comm_init(void) 
{
#ifndef IPSEC_LEVEL_REQUIRE
  fprintf(stderr, "No socket SECURITY: firewall-to-firewall communication disabled\n");
  log_entry("No socket SECURITY: firewall-to-firewall communication disabled");
#else
  struct sockaddr_in myaddr;
  int sockval;

  memset(&myaddr, 0, sizeof(myaddr));
  myaddr.sin_family      = AF_INET;
  myaddr.sin_port        = htons(SERVER_PORT);
  myaddr.sin_addr.s_addr = htonl(INADDR_ANY);

  if ((sockcomm = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("Cannot create socket");
    exit(2);
  }

  if (bind(sockcomm, (struct sockaddr *) &myaddr, sizeof(myaddr)) < 0) {
    perror("Cannot bind address to socket");
    close(sockcomm);
    exit(2);
  }

  sockval = IPSEC_LEVEL_REQUIRE;
  if (setsockopt(sockcomm, SOL_SOCKET, SO_SECURITY_AUTHENTICATION, &sockval, sizeof(sockval)) < 0)
  {
    fprintf(stderr, "No socket SECURITY: firewall-to-firewall communication disabled\n");
    log_entry("No socket SECURITY: firewall-to-firewall communication disabled");
    close(sockcomm);
    sockcomm = -1;
    return;
  }

  myaddr.sin_port = htons(FRIENDS_PORT);

  if ((sockfriends = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("Cannot create socket");
    exit(2);
  }

  if (bind(sockfriends, (struct sockaddr *) &myaddr, sizeof(myaddr)) < 0) {
    perror("Cannot bind address to socket");
    close(sockcomm);
    close(sockfriends);
    exit(2);
  }

  sockval = IPSEC_LEVEL_REQUIRE;
  if (setsockopt(sockfriends, SOL_SOCKET, SO_SECURITY_AUTHENTICATION, &sockval, sizeof(sockval)) < 0)
  {
    fprintf(stderr, "No socket SECURITY: firewall-to-firewall communication disabled\n");
    log_entry("No socket SECURITY: firewall-to-firewall communication disabled");
    close(sockcomm);
    close(sockfriends);
    sockcomm    =
    sockfriends = -1;
  }
#endif
}

